#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Core errors that may be raised by this API implementation."""
__all__ = [
    "HikariError",
    "NotFoundHTTPError",
    "UnauthorizedHTTPError",
    "BadRequestHTTPError",
    "ClientHTTPError",
    "ServerHTTPError",
    "CodedHTTPError",
    "HTTPError",
    "GatewayZombiedError",
    "GatewayNeedsShardingError",
    "GatewayMustReconnectError",
    "GatewayInvalidSessionError",
    "GatewayInvalidTokenError",
    "GatewayServerClosedConnectionError",
    "GatewayClientClosedError",
    "GatewayClientDisconnectedError",
    "GatewayError",
]

import typing

from hikari.net import codes
from hikari.net import routes


class HikariError(RuntimeError):
    """Base for an error raised by this API.

    Any should derive from this.

    Note
    ----
    You should never initialize this exception directly.
    """

    __slots__ = ()


class GatewayError(HikariError):
    """A base exception type for anything that can be thrown by the Gateway.

    Parameters
    ----------
    reason : :obj:`~str`
        A string explaining the issue.
    """

    #: A string to explain the issue.
    #:
    #: :type: :obj:`~str`
    reason: str

    def __init__(self, reason: str) -> None:
        super().__init__()
        self.reason = reason

    def __str__(self) -> str:
        return self.reason


class GatewayClientClosedError(GatewayError):
    """An exception raised when you programmatically shut down the bot client-side.

    Parameters
    ----------
    reason : :obj:`~str`
        A string explaining the issue.
    """

    def __init__(self, reason: str = "The gateway client has been closed") -> None:
        super().__init__(reason)


class GatewayClientDisconnectedError(GatewayError):
    """An exception raised when the bot client-side disconnects unexpectedly.

    Parameters
    ----------
    reason : :obj:`~str`
        A string explaining the issue.
    """

    def __init__(self, reason: str = "The gateway client has disconnected unexpectedly") -> None:
        super().__init__(reason)


class GatewayServerClosedConnectionError(GatewayError):
    """An exception raised when the server closes the connection.

    Parameters
    ----------
    close_code : :obj:`~hikari.net.codes.GatewayCloseCode`, :obj:`~int`, optional
        The close code provided by the server, if there was one.
    reason : :obj:`~str`, optional
        A string explaining the issue.
    """

    close_code: typing.Union[codes.GatewayCloseCode, int, None]

    def __init__(
        self,
        close_code: typing.Optional[typing.Union[codes.GatewayCloseCode, int]] = None,
        reason: typing.Optional[str] = None,
    ) -> None:
        try:
            name = close_code.name
        except AttributeError:
            name = str(close_code) if close_code is not None else "no reason"

        if reason is None:
            reason = f"Gateway connection closed by server ({name})"

        self.close_code = close_code
        super().__init__(reason)


class GatewayInvalidTokenError(GatewayServerClosedConnectionError):
    """An exception that is raised if you failed to authenticate with a valid token to the Gateway."""

    def __init__(self) -> None:
        super().__init__(
            codes.GatewayCloseCode.AUTHENTICATION_FAILED,
            "The account token specified is invalid for the gateway connection",
        )


class GatewayInvalidSessionError(GatewayServerClosedConnectionError):
    """An exception raised if a Gateway session becomes invalid.

    Parameters
    ----------
    can_resume : :obj:`~bool`
        :obj:`~True` if the connection will be able to RESUME next time it starts
        rather than re-IDENTIFYing, or :obj:`~False` if you need to IDENTIFY
        again instead.
    """

    #: :obj:`~True` if the next reconnection can be RESUMED. :obj:`~False` if it
    #: has to be coordinated by re-IDENFITYing.
    #:
    #: :type: :obj:`~bool`
    can_resume: bool

    def __init__(self, can_resume: bool) -> None:
        self.can_resume = can_resume
        instruction = "restart the shard and RESUME" if can_resume else "restart the shard with a fresh session"
        super().__init__(reason=f"The session has been invalidated; {instruction}")


class GatewayMustReconnectError(GatewayServerClosedConnectionError):
    """An exception raised when the Gateway has to re-connect with a new session.

    This will cause a re-IDENTIFY.
    """

    def __init__(self) -> None:
        super().__init__(reason="The gateway server has requested that the client reconnects with a new session")


class GatewayNeedsShardingError(GatewayServerClosedConnectionError):
    """An exception raised if you have too many guilds on one of the current Gateway shards.

    This is a sign you need to increase the number of shards that your bot is
    running with in order to connect to Discord.
    """

    def __init__(self) -> None:
        super().__init__(
            codes.GatewayCloseCode.SHARDING_REQUIRED, "You are in too many guilds. Shard the bot to connect",
        )


class GatewayZombiedError(GatewayClientClosedError):
    """An exception raised if a shard becomes zombied.

    This means that Discord is no longer responding to us, and we have
    disconnected due to a timeout.
    """

    def __init__(self) -> None:
        super().__init__("No heartbeat was received, the connection has been closed")


class HTTPError(HikariError):
    """Base exception raised if an HTTP error occurs.

    Parameters
    ----------
    reason : :obj:`~str`
        A meaningful explanation of the problem.
    """

    #: A meaningful explanation of the problem.
    #:
    #: :type: :obj:`~str`
    reason: str

    def __init__(self, reason: str) -> None:
        super().__init__()
        self.reason = reason

    def __str__(self) -> str:
        return self.reason


class CodedHTTPError(HTTPError):
    """An HTTP exception that has contextual response information with it.

    Parameters
    ----------
    status : :obj:`~int` or :obj:`~hikari.net.codes.HTTPStatusCode`
        The HTTP status code that was returned by the server.
    route : :obj:`~hikari.net.routes.CompiledRoute`
        The HTTP route that was being invoked when this exception occurred.
    message : :obj:`~str`, optional
        An optional message if provided in the response payload.
    json_code : :obj:`~hikari.net.codes.JSONErrorCode`, :obj:`~int`, optional
        An optional error code the server provided us.
    """

    #: The HTTP status code that was returned by the server.
    #:
    #: :type: :obj:`~int` or :obj:`~hikari.net.codes.HTTPStatusCode`
    status: typing.Union[int, codes.HTTPStatusCode]

    #: The HTTP route that was being invoked when this exception occurred.
    #:
    #: :type: :obj:`~hikari.net.routes.CompiledRoute`
    route: routes.CompiledRoute

    #: An optional contextual message the server provided us with in the
    #: response body.
    #
    #: :type: :obj:`~str`, optional
    message: typing.Optional[str]

    #: An optional contextual error code the server provided us with in the
    #: response body.
    #
    #: :type: :obj:`~hikari.net.codes.JSONErrorCode` or :obj:`~int`, optional
    json_code: typing.Optional[typing.Union[codes.JSONErrorCode, int]]

    def __init__(
        self,
        status: typing.Union[int, codes.HTTPStatusCode],
        route: routes.CompiledRoute,
        message: typing.Optional[str],
        json_code: typing.Union[codes.JSONErrorCode, int, None],
    ) -> None:
        super().__init__(str(status))
        self.status = status
        self.route = route
        self.message = message
        self.json_code = json_code

    def __str__(self) -> str:
        return f"{self.reason}: ({self.json_code}) {self.message}"


class ServerHTTPError(CodedHTTPError):
    """An exception raised if a server-side error occurs when interacting with the REST API.

    If you get these, DO NOT PANIC! Your bot is working perfectly fine. Discord
    have probably broken something again.
    """


class ClientHTTPError(CodedHTTPError):
    """An exception raised if a server-side error occurs when interacting with the REST API.

    If you get one of these, you most likely have a mistake in your code, or
    have found a bug with this library.

    If you are sure that your code is correct, please register a bug at
    https://gitlab.com/nekokatt/hikari/issues and we will take a look for you.
    """


class BadRequestHTTPError(CodedHTTPError):
    """A specific case of :obj:`~CodedHTTPError`.

    This can occur hat occurs when you send Discord information in an unexpected
    format, miss required information out, or give bad values for stuff.

    An example might be sending a message without any content, or an embed with
    more than 6000 characters.

    Parameters
    ----------
    route : :obj:`~hikari.net.routes.CompiledRoute`
        The HTTP route that was being invoked when this exception occurred.
    message : :obj:`~str`, optional
        An optional message if provided in the response payload.
    json_code : :obj:`~hikari.net.codes.JSONErrorCode`, :obj:`~int`, optional
        An optional error code the server provided us.
    """

    def __init__(
        self,
        route: routes.CompiledRoute,
        message: typing.Optional[str],
        json_code: typing.Optional[typing.Union[codes.JSONErrorCode, int]],
    ) -> None:
        super().__init__(codes.HTTPStatusCode.BAD_REQUEST, route, message, json_code)


class UnauthorizedHTTPError(ClientHTTPError):
    """A specific case of :obj:`~ClientHTTPError`.

    This occurs when you have invalid authorization details to access
    the given resource.

    This usually means that you have an incorrect token.

    Parameters
    ----------
    route : :obj:`~hikari.net.routes.CompiledRoute`
        The HTTP route that was being invoked when this exception occurred.
    message : :obj:`~str`, optional
        An optional message if provided in the response payload.
    json_code : :obj:`~hikari.net.codes.JSONErrorCode`, :obj:`~int`, optional
        An optional error code the server provided us.
    """

    def __init__(
        self,
        route: routes.CompiledRoute,
        message: typing.Optional[str],
        json_code: typing.Optional[typing.Union[codes.JSONErrorCode, int]],
    ) -> None:
        super().__init__(codes.HTTPStatusCode.UNAUTHORIZED, route, message, json_code)


class ForbiddenHTTPError(ClientHTTPError):
    """A specific case of :obj:`~ClientHTTPError`.

    This occurs when you are missing permissions, or are using an endpoint that
    your account is not allowed to see without being whitelisted.

    This will not occur if your token is invalid.

    Parameters
    ----------
    route : :obj:`~hikari.net.routes.CompiledRoute`
        The HTTP route that was being invoked when this exception occurred.
    message : :obj:`~str`, optional
        An optional message if provided in the response payload.
    json_code : :obj:`~hikari.net.codes.JSONErrorCode`, :obj:`~int`, optional
        An optional error code the server provided us.
    """

    def __init__(
        self,
        route: routes.CompiledRoute,
        message: typing.Optional[str],
        json_code: typing.Optional[typing.Union[codes.JSONErrorCode, int]],
    ) -> None:
        super().__init__(codes.HTTPStatusCode.FORBIDDEN, route, message, json_code)


class NotFoundHTTPError(ClientHTTPError):
    """A specific case of :obj:`~ClientHTTPError`.

    This occurs when you try to refer to something that doesn't exist on Discord.
    This might be referring to a user ID, channel ID, guild ID, etc that does
    not exist, or it might be attempting to use an HTTP endpoint that is not
    found.

    Parameters
    ----------
    route : :obj:`~hikari.net.routes.CompiledRoute`
        The HTTP route that was being invoked when this exception occurred.
    message : :obj:`~str`, optional
        An optional message if provided in the response payload.
    json_code : :obj:`~hikari.net.codes.JSONErrorCode`, :obj:`~int`, optional
        An optional error code the server provided us.
    """

    def __init__(
        self,
        route: routes.CompiledRoute,
        message: typing.Optional[str],
        json_code: typing.Optional[typing.Union[codes.JSONErrorCode, int]],
    ) -> None:
        super().__init__(codes.HTTPStatusCode.NOT_FOUND, route, message, json_code)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Shard intents for controlling which events the application receives."""
__all__ = ["Intent"]

import enum

from hikari.internal import more_enums


@enum.unique
class Intent(more_enums.FlagMixin, enum.IntFlag):
    """Represents an intent on the gateway.

    This is a bitfield representation of all the categories of event
    that you wish to receive.

    Any events not in an intent category will be fired regardless of what
    intents you provide.

    Warnings
    --------
    If you are using the V7 Gateway, you will be REQUIRED to provide some form
    of intent value when you connect. Failure to do so may result in immediate
    termination of the session server-side.

    Notes
    -----
    Discord now places limits on certain events you can receive without
    whitelisting your bot first. On the ``Bot`` tab in the developer's portal
    for your bot, you should now have the option to enable functionality
    for receiving these events.

    If you attempt to request an intent type that you have not whitelisted
    your bot for, you will be disconnected on startup with a ``4014`` closure
    code.
    """

    #: Subscribes to the following events:
    #: * GUILD_CREATE
    #: * GUILD_UPDATE
    #: * GUILD_DELETE
    #: * GUILD_ROLE_CREATE
    #: * GUILD_ROLE_UPDATE
    #: * GUILD_ROLE_DELETE
    #: * CHANNEL_CREATE
    #: * CHANNEL_UPDATE
    #: * CHANNEL_DELETE
    #: * CHANNEL_PINS_UPDATE
    GUILDS = 1 << 0

    #: Subscribes to the following events:
    #: * GUILD_MEMBER_ADD
    #: * GUILD_MEMBER_UPDATE
    #: * GUILD_MEMBER_REMOVE
    #:
    #: Warnings
    #: --------
    #: This intent is privileged, and requires enabling/whitelisting to use.
    GUILD_MEMBERS = 1 << 1

    #: Subscribes to the following events:
    #: * GUILD_BAN_ADD
    #: * GUILD_BAN_REMOVE
    GUILD_BANS = 1 << 2

    #: Subscribes to the following events:
    #: * GUILD_EMOJIS_UPDATE
    GUILD_EMOJIS = 1 << 3

    #: Subscribes to the following events:
    #: * GUILD_INTEGRATIONS_UPDATE
    GUILD_INTEGRATIONS = 1 << 4

    #: Subscribes to the following events:
    #: * WEBHOOKS_UPDATE
    GUILD_WEBHOOKS = 1 << 5

    #: Subscribes to the following events:
    #: * INVITE_CREATE
    #: * INVITE_DELETE
    GUILD_INVITES = 1 << 6

    #: Subscribes to the following events:
    #: * VOICE_STATE_UPDATE
    GUILD_VOICE_STATES = 1 << 7

    #: Subscribes to the following events:
    #: * PRESENCE_UPDATE
    #:
    #: Warnings
    #: --------
    #: This intent is privileged, and requires enabling/whitelisting to use.
    GUILD_PRESENCES = 1 << 8

    #: Subscribes to the following events:
    #: * MESSAGE_CREATE
    #: * MESSAGE_UPDATE
    #: * MESSAGE_DELETE
    #: * MESSAGE_BULK
    GUILD_MESSAGES = 1 << 9

    #: Subscribes to the following events:
    #: * MESSAGE_REACTION_ADD
    #: * MESSAGE_REACTION_REMOVE
    #: * MESSAGE_REACTION_REMOVE_ALL
    #: * MESSAGE_REACTION_REMOVE_EMOJI
    GUILD_MESSAGE_REACTIONS = 1 << 10

    #: Subscribes to the following events:
    #: * TYPING_START
    GUILD_MESSAGE_TYPING = 1 << 11

    #: Subscribes to the following events:
    #: * CHANNEL_CREATE
    #: * MESSAGE_CREATE
    #: * MESSAGE_UPDATE
    #: * MESSAGE_DELETE
    DIRECT_MESSAGES = 1 << 12

    #: Subscribes to the following events:
    #: * MESSAGE_REACTION_ADD
    #: * MESSAGE_REACTION_REMOVE
    #: * MESSAGE_REACTION_REMOVE_ALL
    DIRECT_MESSAGE_REACTIONS = 1 << 13

    #: Subscribes to the following events
    #: * TYPING_START
    DIRECT_MESSAGE_TYPING = 1 << 14



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe audit logs on Discord.

.. inheritance-diagram::
    hikari.audit_logs
    :parts: 1
"""
__all__ = [
    "AuditLog",
    "AuditLogChange",
    "AuditLogChangeKey",
    "AuditLogEntry",
    "AuditLogEventType",
    "BaseAuditLogEntryInfo",
    "ChannelOverwriteEntryInfo",
    "get_entry_info_entity",
    "MemberDisconnectEntryInfo",
    "MemberMoveEntryInfo",
    "MemberPruneEntryInfo",
    "MessageBulkDeleteEntryInfo",
    "MessageDeleteEntryInfo",
    "MessagePinEntryInfo",
]

import abc
import copy
import datetime
import enum
import typing

import attr

from hikari import bases
from hikari import channels
from hikari import colors
from hikari import guilds
from hikari import permissions
from hikari import users as _users
from hikari import webhooks as _webhooks
from hikari.internal import conversions
from hikari.internal import marshaller
from hikari.internal import more_collections
from hikari.internal import more_typing


class AuditLogChangeKey(str, enum.Enum):
    """Commonly known and documented keys for audit log change objects.

    Others may exist. These should be expected to default to the raw string
    Discord provided us.
    """

    NAME = "name"
    ICON_HASH = "icon_hash"
    SPLASH_HASH = "splash_hash"
    OWNER_ID = "owner_id"
    REGION = "region"
    AFK_CHANNEL_ID = "afk_channel_id"
    AFK_TIMEOUT = "afk_timeout"
    MFA_LEVEL = "mfa_level"
    VERIFICATION_LEVEL = "verification_level"
    EXPLICIT_CONTENT_FILTER = "explicit_content_filter"
    DEFAULT_MESSAGE_NOTIFICATIONS = "default_message_notifications"
    VANITY_URL_CODE = "vanity_url_code"
    ADD_ROLE_TO_MEMBER = "$add"
    REMOVE_ROLE_FROM_MEMBER = "$remove"
    PRUNE_DELETE_DAYS = "prune_delete_days"
    WIDGET_ENABLED = "widget_enabled"
    WIDGET_CHANNEL_ID = "widget_channel_id"
    POSITION = "position"
    TOPIC = "topic"
    BITRATE = "bitrate"
    PERMISSION_OVERWRITES = "permission_overwrites"
    NSFW = "nsfw"
    APPLICATION_ID = "application_id"
    PERMISSIONS = "permissions"
    COLOR = "color"
    HOIST = "hoist"
    MENTIONABLE = "mentionable"
    ALLOW = "allow"
    DENY = "deny"
    INVITE_CODE = "code"
    CHANNEL_ID = "channel_id"
    INVITER_ID = "inviter_id"
    MAX_USES = "max_uses"
    USES = "uses"
    MAX_AGE = "max_age"
    TEMPORARY = "temporary"
    DEAF = "deaf"
    MUTE = "mute"
    NICK = "nick"
    AVATAR_HASH = "avatar_hash"
    ID = "id"
    TYPE = "type"
    ENABLE_EMOTICONS = "enable_emoticons"
    EXPIRE_BEHAVIOR = "expire_behavior"
    EXPIRE_GRACE_PERIOD = "expire_grace_period"
    RATE_LIMIT_PER_USER = "rate_limit_per_user"
    SYSTEM_CHANNEL_ID = "system_channel_id"

    #: Alias for "COLOR"
    COLOUR = COLOR

    def __str__(self) -> str:
        return self.name

    __repr__ = __str__


AUDIT_LOG_ENTRY_CONVERTERS = {
    AuditLogChangeKey.OWNER_ID: bases.Snowflake.deserialize,
    AuditLogChangeKey.AFK_CHANNEL_ID: bases.Snowflake.deserialize,
    AuditLogChangeKey.AFK_TIMEOUT: lambda payload: datetime.timedelta(seconds=payload),
    AuditLogChangeKey.MFA_LEVEL: guilds.GuildMFALevel,
    AuditLogChangeKey.VERIFICATION_LEVEL: guilds.GuildVerificationLevel,
    AuditLogChangeKey.EXPLICIT_CONTENT_FILTER: guilds.GuildExplicitContentFilterLevel,
    AuditLogChangeKey.DEFAULT_MESSAGE_NOTIFICATIONS: guilds.GuildMessageNotificationsLevel,
    AuditLogChangeKey.ADD_ROLE_TO_MEMBER: lambda payload: {
        role.id: role for role in map(guilds.PartialGuildRole.deserialize, payload)
    },
    AuditLogChangeKey.REMOVE_ROLE_FROM_MEMBER: lambda payload: {
        role.id: role for role in map(guilds.PartialGuildRole.deserialize, payload)
    },
    AuditLogChangeKey.PRUNE_DELETE_DAYS: lambda payload: datetime.timedelta(days=int(payload)),
    AuditLogChangeKey.WIDGET_CHANNEL_ID: bases.Snowflake.deserialize,
    AuditLogChangeKey.POSITION: int,
    AuditLogChangeKey.BITRATE: int,
    AuditLogChangeKey.PERMISSION_OVERWRITES: lambda payload: {
        overwrite.id: overwrite for overwrite in map(channels.PermissionOverwrite.deserialize, payload)
    },
    AuditLogChangeKey.APPLICATION_ID: bases.Snowflake.deserialize,
    AuditLogChangeKey.PERMISSIONS: permissions.Permission,
    AuditLogChangeKey.COLOR: colors.Color,
    AuditLogChangeKey.ALLOW: permissions.Permission,
    AuditLogChangeKey.DENY: permissions.Permission,
    AuditLogChangeKey.CHANNEL_ID: bases.Snowflake.deserialize,
    AuditLogChangeKey.INVITER_ID: bases.Snowflake.deserialize,
    AuditLogChangeKey.MAX_USES: lambda payload: int(payload) if payload > 0 else float("inf"),
    AuditLogChangeKey.USES: int,
    AuditLogChangeKey.MAX_AGE: lambda payload: datetime.timedelta(seconds=payload) if payload > 0 else None,
    AuditLogChangeKey.ID: bases.Snowflake.deserialize,
    AuditLogChangeKey.TYPE: str,
    AuditLogChangeKey.ENABLE_EMOTICONS: bool,
    AuditLogChangeKey.EXPIRE_BEHAVIOR: guilds.IntegrationExpireBehaviour,
    AuditLogChangeKey.EXPIRE_GRACE_PERIOD: lambda payload: datetime.timedelta(days=payload),
    AuditLogChangeKey.RATE_LIMIT_PER_USER: lambda payload: datetime.timedelta(seconds=payload),
    AuditLogChangeKey.SYSTEM_CHANNEL_ID: bases.Snowflake.deserialize,
}


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class AuditLogChange(bases.HikariEntity, marshaller.Deserializable):
    """Represents a change made to an audit log entry's target entity."""

    #: The new value of the key, if something was added or changed.
    #:
    #: :type: :obj:`~typing.Any`, optional
    new_value: typing.Optional[typing.Any] = marshaller.attrib()

    #: The old value of the key, if something was removed or changed.
    #:
    #: :type: :obj:`~typing.Any`, optional
    old_value: typing.Optional[typing.Any] = marshaller.attrib()

    #: The name of the audit log change's key.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~AuditLogChangeKey`, :obj:`~str` ]
    key: typing.Union[AuditLogChangeKey, str] = marshaller.attrib()

    @classmethod
    def deserialize(cls, payload: typing.Mapping[str, str]) -> "AuditLogChange":
        """Deserialize this model from a raw payload."""
        key = conversions.try_cast(payload["key"], AuditLogChangeKey, payload["key"])
        new_value = payload.get("new_value")
        old_value = payload.get("old_value")
        if value_converter := AUDIT_LOG_ENTRY_CONVERTERS.get(key):
            new_value = value_converter(new_value) if new_value is not None else None
            old_value = value_converter(old_value) if old_value is not None else None

        # noinspection PyArgumentList
        return cls(key=key, new_value=new_value, old_value=old_value)


@enum.unique
class AuditLogEventType(enum.IntEnum):
    """The type of event that occurred."""

    GUILD_UPDATE = 1
    CHANNEL_CREATE = 10
    CHANNEL_UPDATE = 11
    CHANNEL_DELETE = 12
    CHANNEL_OVERWRITE_CREATE = 13
    CHANNEL_OVERWRITE_UPDATE = 14
    CHANNEL_OVERWRITE_DELETE = 15
    MEMBER_KICK = 20
    MEMBER_PRUNE = 21
    MEMBER_BAN_ADD = 22
    MEMBER_BAN_REMOVE = 23
    MEMBER_UPDATE = 24
    MEMBER_ROLE_UPDATE = 25
    MEMBER_MOVE = 26
    MEMBER_DISCONNECT = 27
    BOT_ADD = 28
    ROLE_CREATE = 30
    ROLE_UPDATE = 31
    ROLE_DELETE = 32
    INVITE_CREATE = 40
    INVITE_UPDATE = 41
    INVITE_DELETE = 42
    WEBHOOK_CREATE = 50
    WEBHOOK_UPDATE = 51
    WEBHOOK_DELETE = 52
    EMOJI_CREATE = 60
    EMOJI_UPDATE = 61
    EMOJI_DELETE = 62
    MESSAGE_DELETE = 72
    MESSAGE_BULK_DELETE = 73
    MESSAGE_PIN = 74
    MESSAGE_UNPIN = 75
    INTEGRATION_CREATE = 80
    INTEGRATION_UPDATE = 81
    INTEGRATION_DELETE = 82


# Ignore docstring not starting in an imperative mood
def register_audit_log_entry_info(
    type_: AuditLogEventType, *additional_types: AuditLogEventType
) -> typing.Callable[[typing.Type["BaseAuditLogEntryInfo"]], typing.Type["BaseAuditLogEntryInfo"]]:  # noqa: D401
    """Generates a decorator for defined audit log entry info entities.

    Allows them to be associated with given entry type(s).

    Parameters
    ----------
    type_ : :obj:`~AuditLogEventType`
        An entry types to associate the entity with.
    *additional_types : :obj:`~AuditLogEventType`
        Extra entry types to associate the entity with.

    Returns
    -------
    ``decorator(T) -> T``
        The decorator to decorate the class with.
    """

    def decorator(cls):
        mapping = getattr(register_audit_log_entry_info, "types", {})
        for t in [type_, *additional_types]:
            mapping[t] = cls
        setattr(register_audit_log_entry_info, "types", mapping)
        return cls

    return decorator


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class BaseAuditLogEntryInfo(bases.HikariEntity, marshaller.Deserializable, abc.ABC):
    """A base object that all audit log entry info objects will inherit from."""


@register_audit_log_entry_info(
    AuditLogEventType.CHANNEL_OVERWRITE_CREATE,
    AuditLogEventType.CHANNEL_OVERWRITE_UPDATE,
    AuditLogEventType.CHANNEL_OVERWRITE_DELETE,
)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ChannelOverwriteEntryInfo(BaseAuditLogEntryInfo):
    """Represents the extra information for overwrite related audit log entries.

    Will be attached to the overwrite create, update and delete audit log
    entries.
    """

    #: The ID of the overwrite being updated, added or removed (and the entity
    #: it targets).
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The type of entity this overwrite targets.
    #:
    #: :type: :obj:`~hikari.channels.PermissionOverwriteType`
    type: channels.PermissionOverwriteType = marshaller.attrib(deserializer=channels.PermissionOverwriteType)

    #: The name of the role this overwrite targets, if it targets a role.
    #:
    #: :type: :obj:`~str`, optional
    role_name: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)


@register_audit_log_entry_info(AuditLogEventType.MESSAGE_PIN, AuditLogEventType.MESSAGE_UNPIN)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessagePinEntryInfo(BaseAuditLogEntryInfo):
    """The extra information for message pin related audit log entries.

    Will be attached to the message pin and message unpin audit log entries.
    """

    #: The ID of the guild text based channel where this pinned message is
    #: being added or removed.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the message that's being pinned or unpinned.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    message_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)


@register_audit_log_entry_info(AuditLogEventType.MEMBER_PRUNE)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MemberPruneEntryInfo(BaseAuditLogEntryInfo):
    """Represents the extra information attached to guild prune log entries."""

    #: The timedelta of how many days members were pruned for inactivity based
    #: on.
    #:
    #: :type: :obj:`~datetime.timedelta`
    delete_member_days: datetime.timedelta = marshaller.attrib(
        deserializer=lambda payload: datetime.timedelta(days=int(payload))
    )

    #: The number of members who were removed by this prune.
    #:
    #: :type: :obj:`~int`
    members_removed: int = marshaller.attrib(deserializer=int)


@register_audit_log_entry_info(AuditLogEventType.MESSAGE_BULK_DELETE)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageBulkDeleteEntryInfo(BaseAuditLogEntryInfo):
    """Represents extra information for the message bulk delete audit entry."""

    #: The amount of messages that were deleted.
    #:
    #: :type: :obj:`~int`
    count: int = marshaller.attrib(deserializer=int)


@register_audit_log_entry_info(AuditLogEventType.MESSAGE_DELETE)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageDeleteEntryInfo(MessageBulkDeleteEntryInfo):
    """Represents extra information attached to the message delete audit entry."""

    #: The guild text based channel where these message(s) were deleted.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)


@register_audit_log_entry_info(AuditLogEventType.MEMBER_DISCONNECT)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MemberDisconnectEntryInfo(BaseAuditLogEntryInfo):
    """Represents extra information for the voice chat member disconnect entry."""

    #: The amount of members who were disconnected from voice in this entry.
    #:
    #: :type: :obj:`~int`
    count: int = marshaller.attrib(deserializer=int)


@register_audit_log_entry_info(AuditLogEventType.MEMBER_MOVE)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MemberMoveEntryInfo(MemberDisconnectEntryInfo):
    """Represents extra information for the voice chat based member move entry."""

    #: The channel these member(s) were moved to.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)


class UnrecognisedAuditLogEntryInfo(BaseAuditLogEntryInfo):
    """Represents any audit log entry options that haven't been implemented."""

    def __init__(self, payload: typing.Mapping[str, str]) -> None:
        self.__dict__.update(payload)

    @classmethod
    def deserialize(cls, payload: typing.Mapping[str, str]) -> "UnrecognisedAuditLogEntryInfo":
        return cls(payload)


def get_entry_info_entity(type_: int) -> typing.Type[BaseAuditLogEntryInfo]:
    """Get the entity that's registered for an entry's options.

    Parameters
    ----------
    :obj:`~int`
        The int

    Returns
    -------
    :obj:`~typing.Type` [ :obj:`~BaseAuditLogEntryInfo` ]
        The associated options entity. If not implemented then this will be
        :obj:`~UnrecognisedAuditLogEntryInfo`
    """
    types = getattr(register_audit_log_entry_info, "types", more_collections.EMPTY_DICT)
    entry_type = types.get(type_)
    return entry_type if entry_type is not None else UnrecognisedAuditLogEntryInfo


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class AuditLogEntry(bases.UniqueEntity, marshaller.Deserializable):
    """Represents an entry in a guild's audit log."""

    #: The ID of the entity affected by this change, if applicable.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    target_id: typing.Optional[bases.Snowflake] = marshaller.attrib()

    #: A sequence of the changes made to :attr:`target_id`
    #:
    #: :type: :obj:`~typing.Sequence` [ :obj:`~AuditLogChange` ]
    changes: typing.Sequence[AuditLogChange] = marshaller.attrib()

    #: The ID of the user who made this change.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    user_id: bases.Snowflake = marshaller.attrib()

    #: The ID of this entry.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    id: bases.Snowflake = marshaller.attrib()

    #: The type of action this entry represents.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~AuditLogEventType`, :obj:`~str` ]
    action_type: typing.Union[AuditLogEventType, str] = marshaller.attrib()

    #: Extra information about this entry. Will only be provided for certain
    #: :attr:`action_type`.
    #:
    #: :type: :obj:`~BaseAuditLogEntryInfo`, optional
    options: typing.Optional[BaseAuditLogEntryInfo] = marshaller.attrib()

    #: The reason for this change, if set (between 0-512 characters).
    #:
    #: :type: :obj:`~str`
    reason: typing.Optional[str] = marshaller.attrib()

    @classmethod
    def deserialize(cls, payload: typing.Mapping[str, str]) -> "AuditLogEntry":
        """Deserialize this model from a raw payload."""
        action_type = conversions.try_cast(payload["action_type"], AuditLogEventType, payload["action_type"])
        if target_id := payload.get("target_id"):
            target_id = bases.Snowflake.deserialize(target_id)

        if (options := payload.get("options")) is not None:
            if option_converter := get_entry_info_entity(action_type):
                options = option_converter.deserialize(options)

        # noinspection PyArgumentList
        return cls(
            target_id=target_id,
            changes=[
                AuditLogChange.deserialize(payload)
                for payload in payload.get("changes", more_collections.EMPTY_SEQUENCE)
            ],
            user_id=bases.Snowflake.deserialize(payload["user_id"]),
            id=bases.Snowflake.deserialize(payload["id"]),
            action_type=action_type,
            options=options,
            reason=payload.get("reason"),
        )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class AuditLog(bases.HikariEntity, marshaller.Deserializable):
    """Represents a guilds audit log."""

    #: A sequence of the audit log's entries.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~AuditLogEntry` ]
    entries: typing.Mapping[bases.Snowflake, AuditLogEntry] = marshaller.attrib(
        raw_name="audit_log_entries",
        deserializer=lambda payload: {entry.id: entry for entry in map(AuditLogEntry.deserialize, payload)},
    )

    #: A mapping of the partial objects of integrations found in this audit log.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.guilds.GuildIntegration` ]
    integrations: typing.Mapping[bases.Snowflake, guilds.GuildIntegration] = marshaller.attrib(
        deserializer=lambda payload: {
            integration.id: integration for integration in map(guilds.PartialGuildIntegration.deserialize, payload)
        }
    )

    #: A mapping of the objects of users found in this audit log.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.users.User` ]
    users: typing.Mapping[bases.Snowflake, _users.User] = marshaller.attrib(
        deserializer=lambda payload: {user.id: user for user in map(_users.User.deserialize, payload)}
    )

    #: A mapping of the objects of webhooks found in this audit log.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.webhooks.Webhook` ]
    webhooks: typing.Mapping[bases.Snowflake, _webhooks.Webhook] = marshaller.attrib(
        deserializer=lambda payload: {webhook.id: webhook for webhook in map(_webhooks.Webhook.deserialize, payload)}
    )


class AuditLogIterator(typing.AsyncIterator[AuditLogEntry]):
    """An async iterator used for iterating through a guild's audit log entries.

    This returns the audit log entries created before a given entry object/ID or
    from the newest audit log entry to the oldest.

    Parameters
    ----------
    guild_id : :obj:`~str`
        The guild ID to look up.
    request : :obj:`~typing.Callable` [ ``...``, :obj:`~typing.Coroutine` [ :obj:`~typing.Any`, :obj:`~typing.Any`, :obj:`~typing.Any` ] ]
        The session bound function that this iterator should use for making
        Get Guild Audit Log requests.
    user_id : :obj:`~str`
        If specified, the user ID to filter by.
    action_type : :obj:`~int`
        If specified, the action type to look up.
    limit : :obj:`~int`
        If specified, the limit to how many entries this iterator should return
        else unlimited.
    before : :obj:`~str`
        If specified, an entry ID to specify where this iterator's returned
        audit log entries should start .

    Note
    ----
    This iterator's attributes :attr:`integrations`, :attr:`users` and
    :attr:`webhooks` will be filled up as this iterator makes requests to the
    Get Guild Audit Log endpoint with the relevant objects for entities
    referenced by returned entries.
    """

    __slots__ = (
        "_buffer",
        "_front",
        "_kwargs",
        "_limit",
        "_request",
        "integrations",
        "users",
        "webhooks",
    )

    #: A mapping of the partial objects of integrations found in this audit log
    #: so far.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.guilds.GuildIntegration` ]
    integrations: typing.Mapping[bases.Snowflake, guilds.GuildIntegration]

    #: A mapping of the objects of users found in this audit log so far.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.users.User` ]
    users: typing.Mapping[bases.Snowflake, _users.User]

    #: A mapping of the objects of webhooks found in this audit log so far.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.webhooks.Webhook` ]
    webhooks: typing.Mapping[bases.Snowflake, _webhooks.Webhook]

    def __init__(
        self,
        guild_id: str,
        request: typing.Callable[..., more_typing.Coroutine[typing.Any]],
        before: typing.Optional[str] = None,
        user_id: str = ...,
        action_type: int = ...,
        limit: typing.Optional[int] = None,
    ) -> None:
        self._kwargs = {"guild_id": guild_id, "user_id": user_id, "action_type": action_type}
        self._limit = limit
        self._buffer = []
        self._request = request
        self._front = before
        self.users = {}
        self.webhooks = {}
        self.integrations = {}

    def __aiter__(self) -> "AuditLogIterator":
        return self

    async def __anext__(self) -> AuditLogEntry:
        if not self._buffer and self._limit != 0:
            await self._fill()
        try:
            entry = AuditLogEntry.deserialize(self._buffer.pop())
            self._front = str(entry.id)
            return entry
        except IndexError:
            raise StopAsyncIteration

    async def _fill(self) -> None:
        """Retrieve entries before :attr:`_front` and add to :attr:`_buffer`."""
        payload = await self._request(
            **self._kwargs,
            before=self._front if self._front is not None else ...,
            limit=100 if self._limit is None or self._limit > 100 else self._limit,
        )
        if self._limit is not None:
            self._limit -= len(payload["audit_log_entries"])

        # Once the resources has been exhausted, discord will return empty lists.
        payload["audit_log_entries"].reverse()
        self._buffer.extend(payload["audit_log_entries"])
        if users := payload.get("users"):
            self.users = copy.copy(self.users)
            self.users.update({u.id: u for u in map(_users.User.deserialize, users)})
        if webhooks := payload.get("webhooks"):
            self.webhooks = copy.copy(self.webhooks)
            self.webhooks.update({w.id: w for w in map(_webhooks.Webhook.deserialize, webhooks)})
        if integrations := payload.get("integrations"):
            self.integrations = copy.copy(self.integrations)
            self.integrations.update({i.id: i for i in map(guilds.PartialGuildIntegration.deserialize, integrations)})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Marshall wrappings for the REST implementation in :mod:`hikari.net.rest`.

This provides an object-oriented interface for interacting with discord's REST
API.
"""

__all__ = ["RESTClient"]

from hikari.clients import configs
from hikari.clients.rest_clients import channel
from hikari.clients.rest_clients import me
from hikari.clients.rest_clients import gateway
from hikari.clients.rest_clients import guild
from hikari.clients.rest_clients import invite
from hikari.clients.rest_clients import oauth2
from hikari.clients.rest_clients import react
from hikari.clients.rest_clients import user
from hikari.clients.rest_clients import voice
from hikari.clients.rest_clients import webhook
from hikari.net import rest_sessions


class RESTClient(
    channel.RESTChannelComponent,
    me.RESTCurrentUserComponent,
    gateway.RESTGatewayComponent,
    guild.RESTGuildComponent,
    invite.RESTInviteComponent,
    oauth2.RESTOAuth2Component,
    react.RESTReactionComponent,
    user.RESTUserComponent,
    voice.RESTVoiceComponent,
    webhook.RESTWebhookComponent,
):
    """
    A marshalling object-oriented REST API client.

    This client bridges the basic REST API exposed by
    :obj:`~hikari.net.rest_sessions.LowLevelRestfulClient` and wraps it in a unit of
    processing that can handle parsing API objects into Hikari entity objects.

    Parameters
    ----------
    config : :obj:`~hikari.clients.configs.RESTConfig`
        A HTTP configuration object.

    Note
    ----
    For all endpoints where a ``reason`` argument is provided, this may be a
    string inclusively between ``0`` and ``512`` characters length, with any
    additional characters being cut off.
    """

    def __init__(self, config: configs.RESTConfig) -> None:
        super().__init__(
            rest_sessions.LowLevelRestfulClient(
                allow_redirects=config.allow_redirects,
                connector=config.tcp_connector,
                proxy_headers=config.proxy_headers,
                proxy_auth=config.proxy_auth,
                ssl_context=config.ssl_context,
                verify_ssl=config.verify_ssl,
                timeout=config.request_timeout,
                token=f"{config.token_type} {config.token}" if config.token_type is not None else config.token,
                version=config.rest_version,
            )
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Network components for the Hikari Discord API.

These components describe the low level parts of Hikari. No model classes exist
for these; the majority of communication is done via JSON arrays and objects.
"""
from hikari.net import codes
from hikari.net import ratelimits
from hikari.net import rest_sessions
from hikari.net import routes
from hikari.net import shards
from hikari.net import user_agents

from hikari.net.codes import *
from hikari.net.rest_sessions import *
from hikari.net.shards import *

__all__ = codes.__all__ + shards.__all__ + rest_sessions.__all__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe Discord gateway events."""

__all__ = [
    "HikariEvent",
    "ExceptionEvent",
    "ConnectedEvent",
    "DisconnectedEvent",
    "StartingEvent",
    "StartedEvent",
    "StoppingEvent",
    "StoppedEvent",
    "ReadyEvent",
    "ResumedEvent",
    "ChannelCreateEvent",
    "ChannelUpdateEvent",
    "ChannelDeleteEvent",
    "ChannelPinUpdateEvent",
    "GuildCreateEvent",
    "GuildUpdateEvent",
    "GuildLeaveEvent",
    "GuildUnavailableEvent",
    "GuildBanAddEvent",
    "GuildBanRemoveEvent",
    "GuildEmojisUpdateEvent",
    "GuildIntegrationsUpdateEvent",
    "GuildMemberAddEvent",
    "GuildMemberUpdateEvent",
    "GuildMemberRemoveEvent",
    "GuildRoleCreateEvent",
    "GuildRoleUpdateEvent",
    "GuildRoleDeleteEvent",
    "InviteCreateEvent",
    "InviteDeleteEvent",
    "MessageCreateEvent",
    "MessageUpdateEvent",
    "MessageDeleteEvent",
    "MessageDeleteBulkEvent",
    "MessageReactionAddEvent",
    "MessageReactionRemoveEvent",
    "MessageReactionRemoveAllEvent",
    "MessageReactionRemoveEmojiEvent",
    "PresenceUpdateEvent",
    "TypingStartEvent",
    "UserUpdateEvent",
    "VoiceStateUpdateEvent",
    "VoiceServerUpdateEvent",
    "WebhookUpdateEvent",
]

import datetime
import typing

import attr

from hikari import bases
from hikari import channels
from hikari import embeds as _embeds
from hikari import emojis as _emojis
from hikari import guilds
from hikari import invites
from hikari import messages
from hikari import applications
from hikari import unset
from hikari import users
from hikari import voices
from hikari.clients import shard_clients
from hikari.internal import conversions
from hikari.internal import marshaller

T_contra = typing.TypeVar("T_contra", contravariant=True)


# Base event, is not deserialized
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class HikariEvent(bases.HikariEntity):
    """The base class that all events inherit from."""


# Synthetic event, is not deserialized, and is produced by the dispatcher.
@attr.attrs(slots=True, auto_attribs=True)
class ExceptionEvent(HikariEvent):
    """Descriptor for an exception thrown while processing an event."""

    #: The exception that was raised.
    #:
    #: :type: :obj:`~Exception`
    exception: Exception

    #: The event that was being invoked when the exception occurred.
    #:
    #: :type: :obj:`~HikariEvent`
    event: HikariEvent

    #: The event that was being invoked when the exception occurred.
    #:
    #: :type: coroutine function ( :obj:`~HikariEvent` ) -> :obj:`~None`
    callback: typing.Callable[[HikariEvent], typing.Awaitable[None]]


# Synthetic event, is not deserialized
@attr.attrs(slots=True, auto_attribs=True)
class StartingEvent(HikariEvent):
    """Event that is fired before the gateway client starts all shards."""


# Synthetic event, is not deserialized
@attr.attrs(slots=True, auto_attribs=True)
class StartedEvent(HikariEvent):
    """Event that is fired when the gateway client starts all shards."""


# Synthetic event, is not deserialized
@attr.attrs(slots=True, auto_attribs=True)
class StoppingEvent(HikariEvent):
    """Event that is fired when the gateway client is instructed to disconnect all shards."""


# Synthetic event, is not deserialized
@attr.attrs(slots=True, auto_attribs=True)
class StoppedEvent(HikariEvent):
    """Event that is fired when the gateway client has finished disconnecting all shards."""


@attr.s(slots=True, kw_only=True, auto_attribs=True)
class ConnectedEvent(HikariEvent, marshaller.Deserializable):
    """Event invoked each time a shard connects."""

    #: The shard that connected.
    #:
    #: :type: :obj:`~hikari.clients.shard_clients.ShardClient`
    shard: shard_clients.ShardClient


@attr.s(slots=True, kw_only=True, auto_attribs=True)
class DisconnectedEvent(HikariEvent, marshaller.Deserializable):
    """Event invoked each time a shard disconnects."""

    #: The shard that disconnected.
    #:
    #: :type: :obj:`~hikari.clients.shard_clients.ShardClient`
    shard: shard_clients.ShardClient


@attr.s(slots=True, kw_only=True, auto_attribs=True)
class ResumedEvent(HikariEvent):
    """Represents a gateway Resume event."""

    #: The shard that reconnected.
    #:
    #: :type: :obj:`~hikari.clients.shard_clients.ShardClient`
    shard: shard_clients.ShardClient


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ReadyEvent(HikariEvent, marshaller.Deserializable):
    """Represents the gateway Ready event.

    This is received only when IDENTIFYing with the gateway.
    """

    #: The gateway version this is currently connected to.
    #:
    #: :type: :obj:`~int`
    gateway_version: int = marshaller.attrib(raw_name="v", deserializer=int)

    #: The object of the current bot account this connection is for.
    #:
    #: :type: :obj:`~hikari.users.MyUser`
    my_user: users.MyUser = marshaller.attrib(raw_name="user", deserializer=users.MyUser.deserialize)

    #: A mapping of the guilds this bot is currently in. All guilds will start
    #: off "unavailable".
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.guilds.UnavailableGuild` ]
    unavailable_guilds: typing.Mapping[bases.Snowflake, guilds.UnavailableGuild] = marshaller.attrib(
        raw_name="guilds",
        deserializer=lambda guilds_objs: {g.id: g for g in map(guilds.UnavailableGuild.deserialize, guilds_objs)},
    )

    #: The id of the current gateway session, used for reconnecting.
    #:
    #: :type: :obj:`~str`
    session_id: str = marshaller.attrib(deserializer=str)

    #: Information about the current shard, only provided when IDENTIFYing.
    #:
    #: :type: :obj:`~typing.Tuple` [ :obj:`~int`, :obj:`~int` ], optional
    _shard_information: typing.Optional[typing.Tuple[int, int]] = marshaller.attrib(
        raw_name="shard", deserializer=tuple, if_undefined=None, default=None
    )

    @property
    def shard_id(self) -> typing.Optional[int]:
        """Zero-indexed ID of the current shard.

        This is only available if this ready event was received while IDENTIFYing.
        """
        return self._shard_information[0] if self._shard_information else None

    @property
    def shard_count(self) -> typing.Optional[int]:
        """Total shard count for this bot.

        This is only available if this ready event was received while IDENTIFYing.
        """
        return self._shard_information[1] if self._shard_information else None


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class BaseChannelEvent(HikariEvent, bases.UniqueEntity, marshaller.Deserializable):
    """A base object that Channel events will inherit from."""

    #: The channel's type.
    #:
    #: :type: :obj:`~hikari.channels.ChannelType`
    type: channels.ChannelType = marshaller.attrib(deserializer=channels.ChannelType)

    #: The ID of the guild this channel is in, will be :obj:`~None` for DMs.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )

    #: The sorting position of this channel, will be relative to the
    #: :attr:`parent_id` if set.
    #:
    #: :type: :obj:`~int`, optional
    position: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)

    #: An mapping of the set permission overwrites for this channel, if applicable.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.channels.PermissionOverwrite` ], optional
    permission_overwrites: typing.Optional[
        typing.Mapping[bases.Snowflake, channels.PermissionOverwrite]
    ] = marshaller.attrib(
        deserializer=lambda overwrites: {o.id: o for o in map(channels.PermissionOverwrite.deserialize, overwrites)},
        if_undefined=None,
        default=None,
    )

    #: The name of this channel, if applicable.
    #:
    #: :type: :obj:`~str`, optional
    name: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)

    #: The topic of this channel, if applicable and set.
    #:
    #: :type: :obj:`~str`, optional
    topic: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None, default=None)

    #: Whether this channel is nsfw, will be :obj:`~None` if not applicable.
    #:
    #: :type: :obj:`~bool`, optional
    is_nsfw: typing.Optional[bool] = marshaller.attrib(
        raw_name="nsfw", deserializer=bool, if_undefined=None, default=None
    )

    #: The ID of the last message sent, if it's a text type channel.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    last_message_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_none=None, if_undefined=None, default=None
    )

    #: The bitrate (in bits) of this channel, if it's a guild voice channel.
    #:
    #: :type: :obj:`~bool`, optional
    bitrate: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)

    #: The user limit for this channel if it's a guild voice channel.
    #:
    #: :type: :obj:`~bool`, optional
    user_limit: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)

    #: The rate limit a user has to wait before sending another message in this
    #: channel, if it's a guild text like channel.
    #:
    #: :type: :obj:`~datetime.timedelta`, optional
    rate_limit_per_user: typing.Optional[datetime.timedelta] = marshaller.attrib(
        deserializer=lambda delta: datetime.timedelta(seconds=delta), if_undefined=None, default=None
    )

    #: A mapping of this channel's recipient users, if it's a DM or group DM.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.users.User` ], optional
    recipients: typing.Optional[typing.Mapping[bases.Snowflake, users.User]] = marshaller.attrib(
        deserializer=lambda recipients: {user.id: user for user in map(users.User.deserialize, recipients)},
        if_undefined=None,
        default=None,
    )

    #: The hash of this channel's icon, if it's a group DM channel and is set.
    #:
    #: :type: :obj:`~str`, optional
    icon_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="icon", deserializer=str, if_undefined=None, if_none=None, default=None
    )

    #: The ID of this channel's creator, if it's a DM channel.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    owner_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )

    #: The ID of the application that created the group DM, if it's a
    #: bot based group DM.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    application_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )

    #: The ID of this channels's parent category within guild, if set.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    parent_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, if_none=None, default=None
    )

    #: The datetime of when the last message was pinned in this channel,
    #: if set and applicable.
    #:
    #: :type: :obj:`~datetime.datetime`, optional
    last_pin_timestamp: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_undefined=None, default=None
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ChannelCreateEvent(BaseChannelEvent):
    """Represents Channel Create gateway events.

    Will be sent when a guild channel is created and before all Create Message
    events that originate from a DM channel.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ChannelUpdateEvent(BaseChannelEvent):
    """Represents Channel Update gateway events."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ChannelDeleteEvent(BaseChannelEvent):
    """Represents Channel Delete gateway events."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ChannelPinUpdateEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent the Channel Pins Update gateway event.

    Sent when a message is pinned or unpinned in a channel but not
    when a pinned message is deleted.
    """

    #: The ID of the guild where this event happened.
    #: Will be :obj:`~None` if this happened in a DM channel.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )

    #: The ID of the channel where the message was pinned or unpinned.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The datetime of when the most recent message was pinned in this channel.
    #: Will be :obj:`~None` if there are no messages pinned after this change.
    #:
    #: :type: :obj:`~datetime.datetime`, optional
    last_pin_timestamp: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_undefined=None, default=None
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildCreateEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Guild Create gateway events.

    Will be received when the bot joins a guild, and when a guild becomes
    available to a guild (either due to outage or at startup).
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildUpdateEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Guild Update gateway events."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildLeaveEvent(HikariEvent, bases.UniqueEntity, marshaller.Deserializable):
    """Fired when the current user leaves the guild or is kicked/banned from it.

    Notes
    -----
    This is fired based on Discord's Guild Delete gateway event.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildUnavailableEvent(HikariEvent, bases.UniqueEntity, marshaller.Deserializable):
    """Fired when a guild becomes temporarily unavailable due to an outage.

    Notes
    -----
    This is fired based on Discord's Guild Delete gateway event.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class BaseGuildBanEvent(HikariEvent, marshaller.Deserializable):
    """A base object that guild ban events will inherit from."""

    #: The ID of the guild this ban is in.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The object of the user this ban targets.
    #:
    #: :type: :obj:`~hikari.users.User`
    user: users.User = marshaller.attrib(deserializer=users.User.deserialize)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildBanAddEvent(BaseGuildBanEvent):
    """Used to represent a Guild Ban Add gateway event."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildBanRemoveEvent(BaseGuildBanEvent):
    """Used to represent a Guild Ban Remove gateway event."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildEmojisUpdateEvent(HikariEvent, marshaller.Deserializable):
    """Represents a Guild Emoji Update gateway event."""

    #: The ID of the guild this emoji was updated in.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The updated mapping of emojis by their ID.
    #:
    #: :type: :obj:`~typing.Mapping` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.emojis.GuildEmoji` ]
    emojis: typing.Mapping[bases.Snowflake, _emojis.GuildEmoji] = marshaller.attrib(
        deserializer=lambda ems: {emoji.id: emoji for emoji in map(_emojis.GuildEmoji.deserialize, ems)}
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildIntegrationsUpdateEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Guild Integration Update gateway events."""

    #: The ID of the guild the integration was updated in.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildMemberAddEvent(HikariEvent, guilds.GuildMember):
    """Used to represent a Guild Member Add gateway event."""

    #: The ID of the guild where this member was added.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildMemberUpdateEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent a Guild Member Update gateway event.

    Sent when a guild member or their inner user object is updated.
    """

    #: The ID of the guild this member was updated in.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: A sequence of the IDs of the member's current roles.
    #:
    #: :type: :obj:`~typing.Sequence` [ :obj:`~hikari.entities.Snowflake` ]
    role_ids: typing.Sequence[bases.Snowflake] = marshaller.attrib(
        raw_name="roles", deserializer=lambda role_ids: [bases.Snowflake.deserialize(rid) for rid in role_ids],
    )

    #: The object of the user who was updated.
    #:
    #:  :type: :obj:`~hikari.users.User`
    user: users.User = marshaller.attrib(deserializer=users.User.deserialize)

    #: This member's nickname. When set to :obj:`~None`, this has been removed
    #: and when set to :obj:`~hikari.unset.UNSET` this hasn't been acted on.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~str`, :obj:`~hikari.unset.UNSET` ], optional
    nickname: typing.Union[None, str, unset.Unset] = marshaller.attrib(
        raw_name="nick", deserializer=str, if_none=None, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The datetime of when this member started "boosting" this guild.
    #: Will be :obj:`~None` if they aren't boosting.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~datetime.datetime`, :obj:`~hikari.unset.UNSET` ], optional
    premium_since: typing.Union[None, datetime.datetime, unset.Unset] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_none=None, if_undefined=unset.Unset, default=unset.UNSET
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildMemberRemoveEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Guild Member Remove gateway events.

    Sent when a member is kicked, banned or leaves a guild.
    """

    #: The ID of the guild this user was removed from.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The object of the user who was removed from this guild.
    #:
    #:  :type: :obj:`~hikari.users.User`
    user: users.User = marshaller.attrib(deserializer=users.User.deserialize)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildRoleCreateEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent a Guild Role Create gateway event."""

    #: The ID of the guild where this role was created.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The object of the role that was created.
    #:
    #: :type: :obj:`~hikari.guilds.GuildRole`
    role: guilds.GuildRole = marshaller.attrib(deserializer=guilds.GuildRole.deserialize)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildRoleUpdateEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent a Guild Role Create gateway event."""

    #: The ID of the guild where this role was updated.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The updated role object.
    #:
    #: :type: :obj:`~hikari.guilds.GuildRole`
    role: guilds.GuildRole = marshaller.attrib(deserializer=guilds.GuildRole.deserialize)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildRoleDeleteEvent(HikariEvent, marshaller.Deserializable):
    """Represents a gateway Guild Role Delete Event."""

    #: The ID of the guild where this role is being deleted.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the role being deleted.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    role_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class InviteCreateEvent(HikariEvent, marshaller.Deserializable):
    """Represents a gateway Invite Create event."""

    #: The ID of the channel this invite targets.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The code that identifies this invite
    #:
    #: :type: :obj:`~str`
    code: str = marshaller.attrib(deserializer=str)

    #: The datetime of when this invite was created.
    #:
    #: :type: :obj:`~datetime.datetime`
    created_at: datetime.datetime = marshaller.attrib(deserializer=conversions.parse_iso_8601_ts)

    #: The ID of the guild this invite was created in, if applicable.
    #: Will be :obj:`~None` for group DM invites.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )

    #: The object of the user who created this invite, if applicable.
    #:
    #: :type: :obj:`~hikari.users.User`, optional
    inviter: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_undefined=None, default=None
    )

    #: The timedelta of how long this invite will be valid for.
    #: If set to :obj:`~None` then this is unlimited.
    #:
    #: :type: :obj:`~datetime.timedelta`, optional
    max_age: typing.Optional[datetime.timedelta] = marshaller.attrib(
        deserializer=lambda age: datetime.timedelta(seconds=age) if age > 0 else None,
    )

    #: The limit for how many times this invite can be used before it expires.
    #: If set to ``0``, or infinity (``float("inf")``) then this is unlimited.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~int`, :obj:`~float` ( ``"inf"`` ) ]
    max_uses: typing.Union[int, float] = marshaller.attrib(deserializer=lambda count: count or float("inf"))

    #: The object of the user who this invite targets, if set.
    #:
    #: :type: :obj:`~hikari.users.User`, optional
    target_user: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_undefined=None, default=None
    )

    #: The type of user target this invite is, if applicable.
    #:
    #: :type: :obj:`~hikari.invites.TargetUserType`, optional
    target_user_type: typing.Optional[invites.TargetUserType] = marshaller.attrib(
        deserializer=invites.TargetUserType, if_undefined=None, default=None
    )

    #: Whether this invite grants temporary membership.
    #:
    #: :type: :obj:`~bool`
    is_temporary: bool = marshaller.attrib(raw_name="temporary", deserializer=bool)

    #: The amount of times this invite has been used.
    #:
    #: :type: :obj:`~int`
    uses: int = marshaller.attrib(deserializer=int)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class InviteDeleteEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Invite Delete gateway events.

    Sent when an invite is deleted for a channel we can access.
    """

    #: The ID of the channel this ID was attached to
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The code of this invite.
    #:
    #: :type: :obj:`~str`
    code: str = marshaller.attrib(deserializer=str)

    #: The ID of the guild this invite was deleted in.
    #: This will be :obj:`~None` if this invite belonged to a DM channel.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageCreateEvent(HikariEvent, messages.Message):
    """Used to represent Message Create gateway events."""


# This is an arbitrarily partial version of `messages.Message`
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageUpdateEvent(HikariEvent, bases.UniqueEntity, marshaller.Deserializable):
    """Represents Message Update gateway events.

    Note
    ----
    All fields on this model except :attr:`channel_id` and :obj:`~`HikariEvent.id`` may be
    set to :obj:`~hikari.unset.UNSET` (a singleton defined in
    ``hikari.entities``) if we have not received information about their
    state from Discord alongside field nullability.
    """

    #: The ID of the channel that the message was sent in.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the guild that the message was sent in.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.unset.UNSET` ]
    guild_id: typing.Union[bases.Snowflake, unset.Unset] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The author of this message.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.users.User`, :obj:`~hikari.unset.UNSET` ]
    author: typing.Union[users.User, unset.Unset] = marshaller.attrib(
        deserializer=users.User.deserialize, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The member properties for the message's author.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.guilds.GuildMember`, :obj:`~hikari.unset.UNSET` ]
    member: typing.Union[guilds.GuildMember, unset.Unset] = marshaller.attrib(
        deserializer=guilds.GuildMember.deserialize, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The content of the message.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~str`, :obj:`~hikari.unset.UNSET` ]
    content: typing.Union[str, unset.Unset] = marshaller.attrib(
        deserializer=str, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The timestamp that the message was sent at.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~datetime.datetime`, :obj:`~hikari.unset.UNSET` ]
    timestamp: typing.Union[datetime.datetime, unset.Unset] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The timestamp that the message was last edited at, or :obj:`~None` if
    #: it wasn't ever edited.
    #:
    #: :type: :obj:`~typing.Union` [  :obj:`~datetime.datetime`, :obj:`~hikari.unset.UNSET` ], optional
    edited_timestamp: typing.Union[datetime.datetime, unset.Unset, None] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_none=None, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: Whether the message is a TTS message.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~bool`, :obj:`~hikari.unset.UNSET` ]
    is_tts: typing.Union[bool, unset.Unset] = marshaller.attrib(
        raw_name="tts", deserializer=bool, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: Whether the message mentions ``@everyone`` or ``@here``.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~bool`, :obj:`~hikari.unset.UNSET` ]
    is_mentioning_everyone: typing.Union[bool, unset.Unset] = marshaller.attrib(
        raw_name="mention_everyone", deserializer=bool, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The users the message mentions.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~typing.Set` [ :obj:`~hikari.entities.Snowflake` ], :obj:`~hikari.unset.UNSET` ]
    user_mentions: typing.Union[typing.Set[bases.Snowflake], unset.Unset] = marshaller.attrib(
        raw_name="mentions",
        deserializer=lambda user_mentions: {bases.Snowflake.deserialize(u["id"]) for u in user_mentions},
        if_undefined=unset.Unset,
        default=unset.UNSET,
    )

    #: The roles the message mentions.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~typing.Set` [ :obj:`~hikari.entities.Snowflake` ], :obj:`~hikari.unset.UNSET` ]
    role_mentions: typing.Union[typing.Set[bases.Snowflake], unset.Unset] = marshaller.attrib(
        raw_name="mention_roles",
        deserializer=lambda role_mentions: {bases.Snowflake.deserialize(r) for r in role_mentions},
        if_undefined=unset.Unset,
        default=unset.UNSET,
    )

    #: The channels the message mentions.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~typing.Set` [ :obj:`~hikari.entities.Snowflake` ], :obj:`~hikari.unset.UNSET` ]
    channel_mentions: typing.Union[typing.Set[bases.Snowflake], unset.Unset] = marshaller.attrib(
        raw_name="mention_channels",
        deserializer=lambda channel_mentions: {bases.Snowflake.deserialize(c["id"]) for c in channel_mentions},
        if_undefined=unset.Unset,
        default=unset.UNSET,
    )

    #: The message attachments.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~typing.Sequence` [ :obj:`~hikari.messages.Attachment` ], :obj:`~hikari.unset.UNSET` ]
    attachments: typing.Union[typing.Sequence[messages.Attachment], unset.Unset] = marshaller.attrib(
        deserializer=lambda attachments: [messages.Attachment.deserialize(a) for a in attachments],
        if_undefined=unset.Unset,
        default=unset.UNSET,
    )

    #: The message's embeds.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~typing.Sequence` [ :obj:`~hikari.embeds.Embed` ], :obj:`~hikari.unset.UNSET` ]
    embeds: typing.Union[typing.Sequence[_embeds.Embed], unset.Unset] = marshaller.attrib(
        deserializer=lambda embed_objs: [_embeds.Embed.deserialize(e) for e in embed_objs],
        if_undefined=unset.Unset,
        default=unset.UNSET,
    )

    #: The message's reactions.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~typing.Sequence` [ :obj:`~hikari.messages.Reaction` ], :obj:`~hikari.unset.UNSET` ]
    reactions: typing.Union[typing.Sequence[messages.Reaction], unset.Unset] = marshaller.attrib(
        deserializer=lambda reactions: [messages.Reaction.deserialize(r) for r in reactions],
        if_undefined=unset.Unset,
        default=unset.UNSET,
    )

    #: Whether the message is pinned.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~bool`, :obj:`~hikari.unset.UNSET` ]
    is_pinned: typing.Union[bool, unset.Unset] = marshaller.attrib(
        raw_name="pinned", deserializer=bool, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: If the message was generated by a webhook, the webhook's id.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.entities.Snowflake`, :obj:`~hikari.unset.UNSET` ]
    webhook_id: typing.Union[bases.Snowflake, unset.Unset] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The message's type.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.messages.MessageType`, :obj:`~hikari.unset.UNSET` ]
    type: typing.Union[messages.MessageType, unset.Unset] = marshaller.attrib(
        deserializer=messages.MessageType, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The message's activity.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.messages.MessageActivity`, :obj:`~hikari.unset.UNSET` ]
    activity: typing.Union[messages.MessageActivity, unset.Unset] = marshaller.attrib(
        deserializer=messages.MessageActivity.deserialize, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The message's application.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.oauth2.Application`, :obj:`~hikari.unset.UNSET` ]
    application: typing.Optional[applications.Application] = marshaller.attrib(
        deserializer=applications.Application.deserialize, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The message's crossposted reference data.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.messages.MessageCrosspost`, :obj:`~hikari.unset.UNSET` ]
    message_reference: typing.Union[messages.MessageCrosspost, unset.Unset] = marshaller.attrib(
        deserializer=messages.MessageCrosspost.deserialize, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The message's flags.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.messages.MessageFlag`, :obj:`~hikari.unset.UNSET` ]
    flags: typing.Union[messages.MessageFlag, unset.Unset] = marshaller.attrib(
        deserializer=messages.MessageFlag, if_undefined=unset.Unset, default=unset.UNSET
    )

    #: The message nonce. This is a string used for validating
    #: a message was sent.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.messages.MessageFlag`, :obj:`~hikari.unset.UNSET` ]
    nonce: typing.Union[str, unset.Unset] = marshaller.attrib(
        deserializer=str, if_undefined=unset.Unset, default=unset.UNSET
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageDeleteEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Message Delete gateway events.

    Sent when a message is deleted in a channel we have access to.
    """

    #: The ID of the channel where this message was deleted.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the guild where this message was deleted.
    #: Will be :obj:`~None` if this message was deleted in a DM channel.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )
    #: The ID of the message that was deleted.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    message_id: bases.Snowflake = marshaller.attrib(raw_name="id", deserializer=bases.Snowflake.deserialize)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageDeleteBulkEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Message Bulk Delete gateway events.

    Sent when multiple messages are deleted in a channel at once.
    """

    #: The ID of the channel these messages have been deleted in.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the channel these messages have been deleted in.
    #: Will be :obj:`~None` if these messages were bulk deleted in a DM channel.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_none=None
    )

    #: A collection of the IDs of the messages that were deleted.
    #:
    #: :type: :obj:`~typing.Set` [ :obj:`~hikari.entities.Snowflake` ]
    message_ids: typing.Set[bases.Snowflake] = marshaller.attrib(
        raw_name="ids", deserializer=lambda msgs: {bases.Snowflake.deserialize(m) for m in msgs}
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageReactionAddEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Message Reaction Add gateway events."""

    #: The ID of the user adding the reaction.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    user_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the channel where this reaction is being added.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the message this reaction is being added to.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    message_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the guild where this reaction is being added, unless this is
    #: happening in a DM channel.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )

    #: The member object of the user who's adding this reaction, if this is
    #: occurring in a guild.
    #:
    #: :type: :obj:`~hikari.guilds.GuildMember`, optional
    member: typing.Optional[guilds.GuildMember] = marshaller.attrib(
        deserializer=guilds.GuildMember.deserialize, if_undefined=None, default=None
    )

    #: The object of the emoji being added.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.emojis.UnknownEmoji`, :obj:`~hikari.emojis.UnicodeEmoji` ]
    emoji: typing.Union[_emojis.UnknownEmoji, _emojis.UnicodeEmoji] = marshaller.attrib(
        deserializer=_emojis.deserialize_reaction_emoji,
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageReactionRemoveEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Message Reaction Remove gateway events."""

    #: The ID of the user who is removing their reaction.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    user_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the channel where this reaction is being removed.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the message this reaction is being removed from.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    message_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the guild where this reaction is being removed, unless this is
    #: happening in a DM channel.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )

    #: The object of the emoji being removed.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.emojis.UnknownEmoji`, :obj:`~hikari.emojis.UnicodeEmoji` ]
    emoji: typing.Union[_emojis.UnicodeEmoji, _emojis.UnknownEmoji] = marshaller.attrib(
        deserializer=_emojis.deserialize_reaction_emoji,
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageReactionRemoveAllEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent Message Reaction Remove All gateway events.

    Sent when all the reactions are removed from a message, regardless of emoji.
    """

    #: The ID of the channel where the targeted message is.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the message all reactions are being removed from.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    message_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the guild where the targeted message is, if applicable.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class MessageReactionRemoveEmojiEvent(HikariEvent, marshaller.Deserializable):
    """Represents Message Reaction Remove Emoji events.

    Sent when all the reactions for a single emoji are removed from a message.
    """

    #: The ID of the channel where the targeted message is.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the guild where the targeted message is, if applicable.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )

    #: The ID of the message the reactions are being removed from.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    message_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The object of the emoji that's being removed.
    #:
    #: :type: :obj:`~typing.Union` [ :obj:`~hikari.emojis.UnknownEmoji`, :obj:`~hikari.emojis.UnicodeEmoji` ]
    emoji: typing.Union[_emojis.UnicodeEmoji, _emojis.UnknownEmoji] = marshaller.attrib(
        deserializer=_emojis.deserialize_reaction_emoji,
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class PresenceUpdateEvent(HikariEvent, guilds.GuildMemberPresence):
    """Used to represent Presence Update gateway events.

    Sent when a guild member changes their presence.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class TypingStartEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent typing start gateway events.

    Received when a user or bot starts "typing" in a channel.
    """

    #: The ID of the channel this typing event is occurring in.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the guild this typing event is occurring in.
    #: Will be :obj:`~None` if this event is happening in a DM channel.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`, optional
    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake.deserialize, if_undefined=None, default=None
    )

    #: The ID of the user who triggered this typing event.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    user_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The datetime of when this typing event started.
    #:
    #: :type: :obj:`~datetime.datetime`
    timestamp: datetime.datetime = marshaller.attrib(
        deserializer=lambda date: datetime.datetime.fromtimestamp(date, datetime.timezone.utc)
    )

    #: The member object of the user who triggered this typing event,
    #: if this was triggered in a guild.
    #:
    #: :type: :obj:`~hikari.guilds.GuildMember`, optional
    member: typing.Optional[guilds.GuildMember] = marshaller.attrib(
        deserializer=guilds.GuildMember.deserialize, if_undefined=None, default=None
    )


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class UserUpdateEvent(HikariEvent, users.MyUser):
    """Used to represent User Update gateway events.

    Sent when the current user is updated.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class VoiceStateUpdateEvent(HikariEvent, voices.VoiceState):
    """Used to represent voice state update gateway events.

    Sent when a user joins, leaves or moves voice channel(s).
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class VoiceServerUpdateEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent voice server update gateway events.

    Sent when initially connecting to voice and when the current voice instance
    falls over to a new server.
    """

    #: The voice connection's token
    #:
    #: :type: :obj:`~str`
    token: str = marshaller.attrib(deserializer=str)

    #: The ID of the guild this voice server update is for
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The uri for this voice server host.
    #:
    #: :type: :obj:`~str`
    endpoint: str = marshaller.attrib(deserializer=str)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class WebhookUpdateEvent(HikariEvent, marshaller.Deserializable):
    """Used to represent webhook update gateway events.

    Sent when a webhook is updated, created or deleted in a guild.
    """

    #: The ID of the guild this webhook is being updated in.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    guild_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)

    #: The ID of the channel this webhook is being updated in.
    #:
    #: :type: :obj:`~hikari.entities.Snowflake`
    channel_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake.deserialize)
